import warnings
from abc import ABC, abstractmethod
from typing import List, Union, Tuple

from numpy import nextafter
from sklearn.base import BaseEstimator

from mindfoundry.optaas.client.constraint import Constraint
from mindfoundry.optaas.client.parameter import GroupParameter, Parameter

SMALLEST_NUMBER_ABOVE_ZERO = nextafter(0.0, 1)
LARGEST_NUMBER_BELOW_ONE = nextafter(1.0, 0)

ParametersAndConstraints = Tuple[List[Parameter], List[Constraint]]


class Optimizable(ABC):
    """Superclass for all optimizable steps."""

    @abstractmethod
    def make_all_parameters_and_constraints(self, estimator_name: str, id_prefix: str,
                                            **kwargs) -> ParametersAndConstraints:
        """Returns all parameters and constraints for multiple estimators in a group or choice."""
        pass


Estimator = Union[BaseEstimator, Optimizable]
EstimatorTuple = Tuple[str, Estimator]


def _get_all_parameters_and_constraints(estimators: List[EstimatorTuple], id_prefix: str,
                                        **kwargs) -> ParametersAndConstraints:
    all_parameters = []
    all_constraints = []

    for estimator_name, estimator in estimators:
        if isinstance(estimator, Optimizable):
            parameters, constraints = estimator.make_all_parameters_and_constraints(estimator_name, id_prefix, **kwargs)
        else:
            warnings.warn(f"{type(estimator).__name__} is not an Optimizable estimator.")
            parameters = [GroupParameter(estimator_name, id=_make_id(id_prefix + estimator_name), items=[])]
            constraints = []

        all_parameters.extend(parameters)
        all_constraints.extend(constraints)

    return all_parameters, all_constraints


def _make_id(estimator_name: str) -> str:
    return estimator_name.replace(' ', '-')


class MissingArgumentError(ValueError):
    """Raised when a required argument is missing from kwargs in :meth:`.OPTaaSClient.create_sklearn_task`"""

    def __init__(self, required_arg: str, estimator) -> None:
        super().__init__(f"{required_arg} is required in kwargs in order to optimize {type(estimator).__name__}")



import warnings
from typing import List, Optional

from requests import Response

from mindfoundry.optaas._version import get_versions


# If the client and server versions differ in the Major or Minor part, we display a warning.
# If they only differ in the Patch, it's ok.

def _get_pieces(version: str) -> List[int]:
    try:
        pieces = version.split('.')
        return [int(pieces[0]), int(pieces[1])]
    except BaseException:
        return []


CLIENT_VERSION = get_versions()['version'].split('.post')[0]
CLIENT_VERSION_PIECES = _get_pieces(CLIENT_VERSION)

NEWER_VERSION_MESSAGE = """
A new version of the OPTaaS client is available. Please run:
    pip install mindfoundry-optaas-client=={}
To stop these messages, use OPTaaSClient(url, api_key, disable_version_check=True)
"""

OLDER_VERSION_MESSAGE = """
Your OPTaaS client version is not in sync with your OPTaaS server. To avoid unexpected behavior, please run:
    pip install mindfoundry-optaas-client=={}
To stop these messages, use OPTaaSClient(url, api_key, disable_version_check=True)
"""


def compare_versions(response: Response) -> None:
    server_version = response.headers.get('X-ApiVersion')
    if server_version:
        server_version = server_version.split('.post')[0]
        if server_version != CLIENT_VERSION:
            server_version_pieces = _get_pieces(server_version)
            warning_message = _get_version_warning_message(server_version_pieces)
            if warning_message:
                warnings.warn(warning_message.format(server_version))


def _get_version_warning_message(server_version_pieces: List[int]) -> Optional[str]:
    for client, server in zip(CLIENT_VERSION_PIECES, server_version_pieces):
        if client < server:
            return NEWER_VERSION_MESSAGE
        if client > server:
            return OLDER_VERSION_MESSAGE
    return None



from typing import List

from sklearn.ensemble import VotingClassifier as BaseVotingClassifier

from mindfoundry.optaas.client.parameter import Parameter, GroupParameter
from mindfoundry.optaas.client.sklearn_pipelines.mixin import OptimizableBaseEstimator
from mindfoundry.optaas.client.sklearn_pipelines.parameter_maker import SklearnParameterMaker
from mindfoundry.optaas.client.sklearn_pipelines.utils import _get_all_parameters_and_constraints, \
    ParametersAndConstraints


class VotingClassifier(BaseVotingClassifier, OptimizableBaseEstimator):
    def make_parameters_and_constraints(self, sk: SklearnParameterMaker, **kwargs) -> ParametersAndConstraints:
        """Generates :class:`Parameters <.Parameter>` and :class:`Constraints <.Constraint>` to optimize a :class:`.VotingClassifier`."""

        grouped_parameters, constraints = _get_all_parameters_and_constraints(self.estimators, sk.prefix, **kwargs)

        parameters: List[Parameter] = []
        for group in grouped_parameters:
            if not isinstance(group, GroupParameter):
                raise ValueError('VotingClassifier does not support Choices')
            for parameter in group.items:
                parameter.name = group.name + '__' + parameter.name
            parameters.extend(group.items)

        parameters.extend([
            sk.CategoricalParameter('voting', values=['hard', 'soft']),
            sk.GroupParameter('weights', optional=True, items=[
                sk.FloatParameter(f'weight_{estimator_name}', minimum=0, maximum=1)
                for estimator_name, _ in self.estimators
            ])
        ])
        return parameters, constraints



from pprint import pformat
from typing import Any, Dict


def _pprint(obj: Any, *keys: str) -> str:
    return pformat({key: getattr(obj, key) for key in keys}, indent=2)


def move_dict_value_up_one_level(dictionary: Dict, key: str) -> None:
    """Changes dict {'key': {'sub_key': 'sub_value'}} to {'key': 'sub_value'}"""
    if key in dictionary:
        dictionary[key] = get_first_value(dictionary[key])


def get_first_key(dictionary: Dict) -> str:
    return list(dictionary.keys())[0]


def get_first_value(dictionary: Dict) -> Any:
    return list(dictionary.values())[0]


get_choice_name = get_first_key  # pylint: disable=invalid-name
get_choice_value = get_first_value  # pylint: disable=invalid-name



from typing import List, Union, TypeVar, Callable, Any, Dict, Sequence

from sklearn.base import BaseEstimator

import mindfoundry.optaas.client.parameter as base
from mindfoundry.optaas.client.sklearn_pipelines.utils import _make_id

T = TypeVar('T', bound=base.Parameter)


class SklearnParameterMaker:
    """Creates :class:`Parameters <.Parameter>` with the correct names and default values for optimizing a sklearn :class:`.Pipeline`

    Convenience methods are provided for each :class:`.Parameter` subclass, so that you can call `sk.IntParameter(...)`
    instead of `IntParameter(...)`.
    """

    def __init__(self, estimator_id: str, estimator: BaseEstimator) -> None:
        self.prefix = estimator_id + '__'
        self.defaults: Dict[str, Any] = estimator.get_params()

    def GroupParameter(self, name: str, items: List[base.Parameter], optional: bool = None) -> base.GroupParameter:  # pylint: disable=invalid-name
        return self.make_parameter(base.GroupParameter, name=name, items=items, optional=optional)

    def ChoiceParameter(self, name: str, choices: List[base.Parameter], optional: bool = None) -> base.ChoiceParameter:  # pylint: disable=invalid-name
        return self.make_parameter(base.ChoiceParameter, name=name, choices=choices, optional=optional)

    def CategoricalParameter(self, name: str, values: Sequence[Union[str, int, float, bool]],  # pylint: disable=invalid-name
                             optional: bool = None) -> base.CategoricalParameter:
        return self.make_parameter(base.CategoricalParameter, name=name, values=values, optional=optional)

    def SubsetParameter(self, name: str, values: List[Union[str, int, float, bool]],  # pylint: disable=invalid-name
                        optional: bool = None) -> base.SubsetParameter:
        return self.make_parameter(base.SubsetParameter, name=name, values=values, optional=optional)

    def ConstantParameter(self, name: str, value: Union[str, int, float, bool],  # pylint: disable=invalid-name
                          optional: bool = None) -> base.ConstantParameter:
        return self.make_parameter(base.ConstantParameter, name=name, value=value, optional=optional, )

    def BoolParameter(self, name: str, optional: bool = None) -> base.BoolParameter:  # pylint: disable=invalid-name
        return self.make_parameter(base.BoolParameter, name=name, optional=optional)

    def IntParameter(self, name: str, minimum: int, maximum: int, distribution: base.Distribution = None,  # pylint: disable=invalid-name
                     optional: bool = None) -> base.IntParameter:
        return self.make_parameter(base.IntParameter, name=name, minimum=minimum, maximum=maximum, optional=optional,
                                   distribution=distribution)

    def FloatParameter(self, name: str, minimum: float, maximum: float, distribution: base.Distribution = None,  # pylint: disable=invalid-name
                       optional: bool = None) -> base.FloatParameter:
        return self.make_parameter(base.FloatParameter, name=name, minimum=minimum, maximum=maximum, optional=optional,
                                   distribution=distribution)

    def IntOrAuto(self, name: str, minimum: int, maximum: int, distribution: base.Distribution = None,  # pylint: disable=invalid-name
                  optional: bool = None) -> base.ChoiceParameter:
        """Creates a choice between an IntParameter and the string 'auto'."""
        return self.IntOrCategorical(name, minimum=minimum, maximum=maximum, categories=["auto"], optional=optional,
                                     distribution=distribution)

    def IntOrCategorical(self, name: str, minimum: int, maximum: int, categories: List[str],  # pylint: disable=invalid-name
                         distribution: base.Distribution = None, optional: bool = None) -> base.ChoiceParameter:
        """Creates a choice between an IntParameter and a CategoricalParameter."""
        return self.ChoiceParameter(name, choices=[
            self.IntParameter(name=name + '_int', minimum=minimum, maximum=maximum, optional=optional,
                              distribution=distribution),
            self.CategoricalParameter(name=name + '_cat', values=categories)
        ])

    def FloatOrAuto(self, name: str, minimum: float, maximum: float, distribution: base.Distribution = None,  # pylint: disable=invalid-name
                    optional: bool = None) -> base.ChoiceParameter:
        """Creates a choice between a FloatParameter and the string 'auto'."""
        return self.FloatOrCategorical(name, minimum=minimum, maximum=maximum, categories=["auto"], optional=optional,
                                       distribution=distribution)

    def FloatOrCategorical(self, name: str, minimum: float, maximum: float, categories: List[str],  # pylint: disable=invalid-name
                           distribution: base.Distribution = None, optional: bool = None) -> base.ChoiceParameter:
        """Creates a choice between a FloatParameter and a CategoricalParameter."""
        return self.ChoiceParameter(name, choices=[
            self.FloatParameter(name=name + '_float', minimum=minimum, maximum=maximum, optional=optional,
                                distribution=distribution),
            self.CategoricalParameter(name=name + '_cat', values=categories)
        ])

    def DictParameter(self, name: str, items: List[base.Parameter], optional: bool = None) -> base.GroupParameter:  # pylint: disable=invalid-name
        """Creates a parameter whose value will be passed into an estimator as a dict."""
        return self.GroupParameter(name, optional=optional, items=[
            self.GroupParameter(name + '_dict_', items=items)
        ])

    def make_parameter(self, parameter_type: Callable[..., T], name: str, **kwargs) -> T:
        """Creates a parameter so as to facilitate the generation of a sklearn Pipeline from a :class:`.Configuration`.

        Args:
            parameter_type (Callable[..., T]): The specific Parameter subclass of the parameter you want to create, e.g. :class:`.IntParameter`.
            name (str): Parameter name, should match the name expected by the estimator's `set_params` method exactly.
            kwargs: Any additional arguments for the parameter constructor, e.g. `minimum`, `maximum`, `choices` etc.
                **Do not** include a value for the `id` and `default` arguments, because it will be overwritten.
                The `id` will be generated from the parameter name (any spaces will be replaced by underscores) and prefixed
                with the estimator name.
                The `default` will be taken from `estimator.get_params()`, i.e. it should be set in the estimator constructor.
        """

        kwargs['name'] = name
        kwargs['id'] = _make_id(self.prefix + name)

        default = self.defaults.get(name)

        if default is None:
            kwargs.pop('default', None)
            if kwargs.get('optional') is True:
                kwargs['include_in_default'] = False
        else:
            if parameter_type == base.ChoiceParameter:
                default_choice = _get_default_choice(default, kwargs)
                if not isinstance(default_choice, base.ConstantParameter):
                    setattr(default_choice, 'default', default)
                if hasattr(default_choice, 'includeInDefault'):
                    delattr(default_choice, 'includeInDefault')
                default = default_choice
            elif parameter_type == base.GroupParameter:
                for i, value in enumerate(default):
                    item = kwargs['items'][i]
                    setattr(item, 'default', value)
                    if hasattr(item, 'includeInDefault'):
                        delattr(item, 'includeInDefault')

            kwargs.pop('include_in_default', None)

            if parameter_type != base.GroupParameter:
                kwargs['default'] = default

        return parameter_type(**kwargs)


def _get_default_choice(default_value, kwargs) -> base.Parameter:
    for choice in kwargs['choices']:
        if choice.is_compatible_value(default_value):
            return choice
    raise ValueError(f"{default_value} is not a valid default value for parameter '{kwargs['name']}'")



from sklearn.decomposition import FastICA as BaseICA

from mindfoundry.optaas.client.constraint import Constraint
from mindfoundry.optaas.client.sklearn_pipelines.mixin import OptimizableBaseEstimator
from mindfoundry.optaas.client.sklearn_pipelines.parameter_maker import SklearnParameterMaker
from mindfoundry.optaas.client.sklearn_pipelines.utils import ParametersAndConstraints


class FastICA(BaseICA, OptimizableBaseEstimator):
    def make_parameters_and_constraints(self, sk: SklearnParameterMaker, **kwargs) -> ParametersAndConstraints:
        """Generates :class:`Parameters <.Parameter>` and :class:`Constraints <.Constraint>` to optimize a :class:`.FastICA` estimator.

        Args:
            feature_count (int): Total number of features in your dataset.
        """

        feature_count = self.get_required_kwarg(kwargs, 'feature_count')
        fun = sk.CategoricalParameter('fun', values=["logcosh", "exp", "cube"])
        alpha = sk.FloatParameter('alpha', minimum=1, maximum=2, optional=True)

        parameters = [
            fun,
            sk.DictParameter('fun_args', items=[alpha]),
            sk.IntParameter('n_components', minimum=1, maximum=feature_count),
            sk.BoolParameter('whiten'),
            sk.CategoricalParameter('algorithm', values=['parallel', 'deflation'])
        ]

        constraints = [
            Constraint(when=fun != 'logcosh', then=alpha.is_absent()),
        ]

        return parameters, constraints



from sklearn.neighbors import KNeighborsClassifier as BaseKNeighborsClassifier, \
    KNeighborsRegressor as BaseKNeighborsRegressor

from mindfoundry.optaas.client.parameter import Distribution
from mindfoundry.optaas.client.sklearn_pipelines.mixin import OptimizableBaseEstimator
from mindfoundry.optaas.client.sklearn_pipelines.parameter_maker import SklearnParameterMaker
from mindfoundry.optaas.client.sklearn_pipelines.utils import ParametersAndConstraints


class _OptimizableKNeighbors(OptimizableBaseEstimator):
    def make_parameters_and_constraints(self, sk: SklearnParameterMaker, **kwargs) -> ParametersAndConstraints:
        """Generates :class:`Parameters <.Parameter>` and :class:`Constraints <.Constraint>` to optimize a KNeighbors estimator."""

        parameters = [
            sk.IntParameter('n_neighbors', minimum=1, maximum=20, distribution=Distribution.LOGUNIFORM),
            sk.CategoricalParameter('weights', values=["uniform", "distance"]),
            sk.CategoricalParameter('p', values=[1, 2]),
        ]

        return parameters, []


class KNeighborsClassifier(BaseKNeighborsClassifier, _OptimizableKNeighbors):
    """Allows us to optimize :class:`.KNeighborsClassifier` estimators."""
    pass


class KNeighborsRegressor(BaseKNeighborsRegressor, _OptimizableKNeighbors):
    """Allows us to optimize :class:`.KNeighborsRegressor` estimators."""
    pass



from typing import List, Any, Dict, TYPE_CHECKING

from mindfoundry.optaas.client.api_key import ApiKey, UserRole
from mindfoundry.optaas.client.constraint import Constraint
from mindfoundry.optaas.client.goal import Goal
from mindfoundry.optaas.client.objective import Objective
from mindfoundry.optaas.client.parameter import Parameter, GroupParameter
from mindfoundry.optaas.client.session import OPTaaSSession, DEFAULT_MAX_RETRIES
from mindfoundry.optaas.client.task import Task

if TYPE_CHECKING:  # pragma: no cover
    from mindfoundry.optaas.client.sklearn_pipelines.mixin import OptimizablePipeline  # pylint: disable=unused-import
    from mindfoundry.optaas.client.sklearn_pipelines.sklearn_task import SklearnTask  # pylint: disable=unused-import

_API_ROOT = '/api/v1'


class OPTaaSClient:
    """Sets up a connection to OPTaaS and allows you to create a :class:`.Task`, retrieve existing Tasks etc.

    Args:
        server_url (str): URL of your OPTaaS server
        api_key (str): Your personal API key
        disable_version_check (bool, default False):
            Set to True if you don't want to be notified when a new version of the client is available
        max_retries (int, default 3): How many times to retry a request if a connection error occurs.
        keep_alive (bool, default True): Very rarely required. Set to False only if you experience connection dropping issues.
    """

    def __init__(self, server_url: str, api_key: str,
                 disable_version_check: bool = False, max_retries: int = DEFAULT_MAX_RETRIES, keep_alive: bool = True) -> None:
        self._session = OPTaaSSession(server_url=server_url, api_key=api_key, disable_version_check=disable_version_check,
                                      max_retries=max_retries, keep_alive=keep_alive)
        root_response = self._session.get(_API_ROOT)
        self._tasks_endpoint = root_response.body['_links']['tasks']['href']
        self._api_key_endpoint = root_response.body['_links']['apiKeys']['href']

    def create_task(self, title: str, parameters: List[Parameter], constraints: List[Constraint] = None,
                    random_seed: int = None, initial_configurations: int = None, objectives: List[Objective] = None,
                    goal: Goal = None, min_known_score: float = None, max_known_score: float = None,
                    user_defined_data: Any = None) -> Task:
        """Creates a new :class:`.Task` by making a POST request to OPTaaS

        Args:
            title (str): Name/description of your Task.
            parameters (List[Parameter]): Parameters that you would like to optimize.
            constraints (List[Constraint]): Constraints on what values can be assigned to Parameters.
            initial_configurations (int, optional, default 10, minimum 1):
                Number of Configurations that OPTaaS will generate upfront. If you are planning to have multiple clients
                working concurrently, set this to be equal to the number of clients.
            objectives (List[Objective], optional): Specification of each objective for a multi-objective optimization.
                Note: Conditional parameters (i.e. ChoiceParameters and anything with optional=True) are not supported
                in multi-objective tasks.
            goal (Goal, optional, default Goal.max): Whether OPTaaS should aim for the lowest (min) or highest (max) score.
                Do not use in multi-objective optimizations (specify goal for each objective instead).
            min_known_score (float, optional): Minimum known score value.
                Do not use in multi-objective optimizations (specify min_known_score for each objective instead).
            max_known_score (float, optional): Maximum known score value.
                Do not use in multi-objective optimizations (specify max_known_score for each objective instead).
            user_defined_data (Any, optional): Any other data you would like to store in the task JSON
            random_seed (int, optional):
                Seed for the random generator used by OPTaaS when generating :class:`Configurations <.Configuration>`.
                If not specified, a new seed will be used for each Task.
                Use this only if you need reproducible results, i.e. if you create 2 Tasks with identical attributes
                including an identical `random_seed`, and you use the same scoring function, then OPTaaS is guaranteed
                to generate the same Configurations in the same order for both Tasks.

        Returns:
            A new :class:`.Task`

        Raises:
            :class:`.OPTaaSError` if the Task data is invalid or the server is unavailable.
        """
        body: Dict = dict(
            title=title,
            parameters=[p.to_json() for p in parameters],
            constraints=[constraint.to_optaas_expression() for constraint in constraints] if constraints else [],
            objectives=[objective.to_json() for objective in objectives] if objectives else None,
            goal=goal.name if goal else None,
            randomSeed=random_seed,
            userDefined=user_defined_data,
            initialConfigurations=initial_configurations,
            minKnownScore=min_known_score,
            maxKnownScore=max_known_score
        )
        body = {key: value for key, value in body.items() if value is not None}
        response = self._session.post(self._tasks_endpoint, body=body)
        return Task(json=response.body, session=self._session)

    def create_sklearn_task(self, title: str, pipeline: 'OptimizablePipeline',
                            additional_parameters: List[Parameter] = None,
                            additional_constraints: List[Constraint] = None,
                            random_seed: int = None, initial_configurations: int = None, objectives: List[Objective] = None,
                            goal: Goal = None, min_known_score: float = None, max_known_score: float = None,
                            user_defined_data: Any = None, **kwargs) -> 'SklearnTask':
        """Creates a new :class:`.SklearnTask` by making a POST request to OPTaaS

        All the arguments from :meth:`.OPTaaSClient.create_task` can be used here except instead of
        `parameters` and `constraints` there is `additional_parameters` and `additional_constraints`.

        Args:
            pipeline (OptimizablePipeline): The pipeline you wish to optimize.
            additional_parameters (List[Parameter], optional):
                Additional parameters that you would like to optimize.
            additional_constraints (List[Constraint], optional):
                Additional constraints on your Parameters.
            kwargs:
                Additional arguments required to optimize certain estimators, e.g. :class:`.PCA` requires `feature_count`.

        Returns:
            A new :class:`.SklearnTask`

        Raises:
            :class:`.MissingArgumentError` if a required argument is missing from `kwargs`.
            :class:`.OPTaaSError` if the Task data is invalid or the server is unavailable.
        """
        from mindfoundry.optaas.client.sklearn_pipelines.sklearn_task import SklearnTask  # pylint: disable=redefined-outer-name

        parameters, constraints = pipeline.make_all_parameters_and_constraints('pipeline', '', **kwargs)

        if additional_parameters:
            parameters.append(GroupParameter('additional', items=additional_parameters))
        if additional_constraints:
            constraints.extend(additional_constraints)

        task = self.create_task(title=title, parameters=parameters, constraints=constraints, objectives=objectives,
                                random_seed=random_seed, goal=goal, initial_configurations=initial_configurations,
                                min_known_score=min_known_score, max_known_score=max_known_score,
                                user_defined_data=user_defined_data)
        return SklearnTask(task, pipeline.estimators)

    def get_all_tasks(self) -> List[Task]:
        """Retrieves a list of all stored Tasks by making a GET request to OPTaaS.

        Returns:
            List of :class:`Tasks <.Task>`

        Raises:
            :class:`.OPTaaSError` if the server is unavailable
        """
        response = self._session.get(self._tasks_endpoint)
        return [Task(json, self._session) for json in response.body['tasks']]

    def get_task(self, task_id: str) -> Task:
        """Retrieves a stored :class:`.Task` by making a GET request to OPTaaS.

        Args:
            task_id (str): unique id for the Task

        Returns:
            A :class:`.Task`

        Raises:
            :class:`.OPTaaSError` if no record is found with the given id or the server is unavailable.
        """
        response = self._session.get(f'{self._tasks_endpoint}/{task_id}')
        return Task(response.body, self._session)

    def get_sklearn_task(self, task_id: str, pipeline: 'OptimizablePipeline') -> 'SklearnTask':
        """Retrieves a stored :class:`.SklearnTask` by making a GET request to OPTaaS.

        This allows you to create a SklearnTask and then use it again in a separate/later session, assuming of course
        that you call this method with the same `estimators` you used to create the original task.

        Args:
            task_id (str): unique id for the Task
            pipeline (OptimizablePipeline): The same pipeline used when calling :meth:`.OPTaaSClient.create_sklearn_task`

        Returns:
            A :class:`.SklearnTask`

        Raises:
            :class:`.OPTaaSError` if no record is found with the given id or the server is unavailable.
        """
        from mindfoundry.optaas.client.sklearn_pipelines.sklearn_task import SklearnTask  # pylint: disable=redefined-outer-name
        task = self.get_task(task_id)
        return SklearnTask(task, pipeline.estimators)

    def get_api_keys(self) -> List[ApiKey]:
        """Retrieves the list of all API keys by making a GET request to OPTaaS. Only available to Admin users."""
        response = self._session.get(self._api_key_endpoint)
        return [ApiKey(json, self._session) for json in response.body['apiKeys']]

    def generate_api_key(self, role: UserRole = None) -> ApiKey:
        """Generates a new API key by making a POST request to OPTaaS. Only available to Admin users."""
        response = self._session.post(self._api_key_endpoint, {} if role is None else {"role": role.value})
        return ApiKey(response.body, self._session)



from sklearn import __version__ as sklearn_version
from sklearn.svm import SVC as BaseSVC, LinearSVC as BaseLinearSVC

from mindfoundry.optaas.client.parameter import Distribution
from mindfoundry.optaas.client.sklearn_pipelines.mixin import OptimizableBaseEstimator
from mindfoundry.optaas.client.sklearn_pipelines.parameter_maker import SklearnParameterMaker
from mindfoundry.optaas.client.sklearn_pipelines.utils import ParametersAndConstraints, SMALLEST_NUMBER_ABOVE_ZERO


class _OptimizableSVC(OptimizableBaseEstimator):
    def make_parameters_and_constraints(self, sk: SklearnParameterMaker, **kwargs) -> ParametersAndConstraints:
        """Generates :class:`Parameters <.Parameter>` and :class:`Constraints <.Constraint>` to optimize an SVC-based estimator."""

        parameters = [
            sk.FloatParameter('C', minimum=0.2, maximum=5.0, distribution=Distribution.LOGUNIFORM),
            sk.FloatParameter('tol', minimum=SMALLEST_NUMBER_ABOVE_ZERO, maximum=1),
            sk.ConstantParameter('class_weight', value='balanced', optional=True),
        ]

        return parameters, []


class SVC(BaseSVC, _OptimizableSVC):
    def make_parameters_and_constraints(self, sk: SklearnParameterMaker, **kwargs) -> ParametersAndConstraints:
        """Generates :class:`Parameters <.Parameter>` and :class:`Constraints <.Constraint>` to optimize an :class:`.SVC` estimator."""

        parameters, constraints = super().make_parameters_and_constraints(sk, **kwargs)

        # Fix hacky default set by sklearn in v0.20
        if sklearn_version.startswith("0.2"):
            gamma_categories = ["scale"]
            if sk.defaults.get("gamma") == "auto_deprecated":
                sk.defaults["gamma"] = "scale"
        else:
            gamma_categories = ["auto"]

        parameters.extend([
            sk.FloatOrCategorical('gamma', minimum=1.0e-5, maximum=1, categories=gamma_categories, distribution=Distribution.LOGUNIFORM),
            sk.CategoricalParameter('kernel', values=['linear', 'poly', 'rbf', 'sigmoid']),
            sk.IntParameter('degree', minimum=2, maximum=10, distribution=Distribution.LOGUNIFORM),
            sk.FloatParameter('coef0', minimum=0, maximum=1),
            sk.BoolParameter('shrinking')
        ])

        return parameters, constraints


class LinearSVC(BaseLinearSVC, _OptimizableSVC):
    pass



from abc import ABC, abstractmethod
from typing import Any, List

from sklearn.base import BaseEstimator

from mindfoundry.optaas.client.parameter import GroupParameter, ChoiceParameter
from mindfoundry.optaas.client.sklearn_pipelines.parameter_maker import SklearnParameterMaker
from mindfoundry.optaas.client.sklearn_pipelines.utils import _make_id, _get_all_parameters_and_constraints, \
    MissingArgumentError, Estimator, EstimatorTuple, ParametersAndConstraints, Optimizable


class OptimizablePipeline(Optimizable):
    """A pipeline-like object to which will be used to generate parameters and constraints for optimization.

    Args:
        estimators (List[EstimatorTuple]):
            List of (name, estimator) tuples as you would provide when creating a sklearn :class:`.Pipeline`.
            An estimator can be:
                * A subclass of :class:`.OptimizableBaseEstimator`
                * A subclass of :class:`.BaseEstimator` (in which case a warning will be displayed informing you that it won't be optimized)
                * An :meth:`~mindfoundry.optaas.client.sklearn_pipelines.mixin.optional_step`
                * A :meth:`~mindfoundry.optaas.client.sklearn_pipelines.mixin.choice` or :meth:`~mindfoundry.optaas.client.sklearn_pipelines.mixin.optional_choice`
                * Another :class:`.OptimizablePipeline`
        optional (bool): Whether this will be an optional step (defaults to False).
    """

    def __init__(self, estimators: List[EstimatorTuple], optional: bool = False) -> None:
        self.estimators = estimators
        self.optional = optional

    def make_all_parameters_and_constraints(self, estimator_name: str, id_prefix: str,
                                            **kwargs) -> ParametersAndConstraints:
        pipeline_name = estimator_name
        prefix = id_prefix + pipeline_name + '__'
        parameters, constraints = _get_all_parameters_and_constraints(self.estimators, prefix, **kwargs)
        grouped_parameter = GroupParameter(pipeline_name, id=_make_id(id_prefix + pipeline_name), items=parameters,
                                           optional=self.optional)
        return [grouped_parameter], constraints


class OptimizableBaseEstimator(BaseEstimator, Optimizable, ABC):
    """Mixin that allows an estimator to be optimized by OPTaaS. Subclasses must implement `make_parameters_and_constraints`."""

    @abstractmethod
    def make_parameters_and_constraints(self, sk: SklearnParameterMaker, **kwargs) -> ParametersAndConstraints:
        """Abstract method that should generate the :class:`Parameters <.Parameter>` and
        :class:`Constraints <.Constraint>` required to optimize a sklearn estimator.

        When implementing this method, make sure to use the :class:`SklearnParameterMaker` `sk` to create parameters
        e.g. call `sk.IntParameter(...)` instead of `IntParameter(...)`.

        If the parameter you want to optimize can take values of different types, use a :meth:`ChoiceParameter <.SklearnParameterMaker.ChoiceParameter>`.
        For an example, see `n_components` in :class:`.PCA`. If the parameter value needs to be a list or array, use a
        :meth:`GroupParameter <.SklearnParameterMaker.GroupParameter>`. For an example, see `weights` in :class:`.VotingClassifier`.

        Args:
            sk (SklearnParameterMaker): Allows you to create parameters with the correct names and defaults.
            kwargs: Additional arguments required to optimize certain estimators, e.g. `feature_count`
                (number of features in your data set, required to optimize :class:`.PCA`)

        Returns:
            A tuple of 2 lists (:class:`Parameters <.Parameter>`, :class:`Constraints <.Constraint>`)

        Raises:
            :class:`.MissingArgumentError` if a required argument is missing from `kwargs`.
        """
        pass

    def get_required_kwarg(self, kwargs, arg_name: str) -> Any:
        """Returns value of a kwarg required to optimize this estimator. Raises error if argument not set.

        Args:
            kwargs: Arguments taken from :meth:`.make_parameters_and_constraints`.
            arg_name (str): Name of the required argument.

        Returns:
            Value of argument.

        Raises:
            :class:`.MissingArgumentError` if the argument is not present.
        """
        if arg_name not in kwargs:
            raise MissingArgumentError(arg_name, self)
        return kwargs[arg_name]

    def make_all_parameters_and_constraints(self, estimator_name: str, id_prefix: str,
                                            **kwargs) -> ParametersAndConstraints:
        estimator_id = _make_id(id_prefix + estimator_name)
        parameter_maker = SklearnParameterMaker(estimator_id, self)
        parameters, constraints = self.make_parameters_and_constraints(parameter_maker, **kwargs)
        grouped_parameter = GroupParameter(estimator_name, id=estimator_id, items=parameters,
                                           optional=isinstance(self, OptionalStepMixin))
        return [grouped_parameter], constraints


class OptionalStepMixin(OptimizableBaseEstimator):
    """Mixin that allows an estimator to be optional, i.e. it may be omitted from a :class:`.Configuration` generated by OPTaaS.

    Example:
        `class MyEstimator(OptionalStepMixin):`

    Your estimator can define `make_parameters_and_constraints` if you wish to optimize its parameters, or you can
    leave it undefined and use the default provided below.
    """

    def make_parameters_and_constraints(self, sk: SklearnParameterMaker, **kwargs) -> ParametersAndConstraints:
        """A default implementation for when you need an optional estimator without optimizing any of its parameters."""
        return [], []


class EstimatorChoice(Optimizable):
    """Allows OPTaaS to choose one of many estimators for a step in a pipeline.

    Args:
        *estimators (Estimator): Estimators from which to choose.
        optional (bool): Whether this will be an optional step (defaults to False).
    """

    def __init__(self, *estimators: Estimator, optional: bool = False) -> None:
        self.estimators = estimators
        self.optional = optional

    def make_all_parameters_and_constraints(self, estimator_name: str, id_prefix: str,
                                            **kwargs) -> ParametersAndConstraints:
        choices, constraints = _get_all_parameters_and_constraints([
            (str(i), estimator) for i, estimator in enumerate(self.estimators)
        ], id_prefix + estimator_name + '__', **kwargs)
        choice_parameter = ChoiceParameter(estimator_name, id=_make_id(estimator_name), choices=choices,
                                           optional=self.optional)
        return [choice_parameter], constraints


def optional_step(estimator: Estimator) -> OptionalStepMixin:
    """Wrapper method to easily make an estimator optional in an OPTaaS :class:`.SklearnTask`.

    The :class:`.OptionalStepMixin` class will be added to the estimator object's base classes (only this instance will
    be affected, not the entire class).

    Example:
        `create_sklearn_task(estimators=[ ('my_optional_step', optional_step(MyEstimator())) ])`
    """

    estimator_type = type(estimator)
    estimator.__class__ = type(estimator_type.__name__, (estimator_type, OptionalStepMixin), {})
    return estimator  # type: ignore


def choice(*estimators: Estimator) -> EstimatorChoice:
    """Convenience method for creating a choice of estimators in a pipeline."""
    return EstimatorChoice(*estimators)


def optional_choice(*estimators: Estimator) -> EstimatorChoice:
    """Convenience method for creating a choice of estimators as an optional step in a pipeline."""
    return EstimatorChoice(*estimators, optional=True)



from sklearn.linear_model import Lasso as BaseLasso, Ridge as BaseRidge

from mindfoundry.optaas.client.sklearn_pipelines.mixin import OptimizableBaseEstimator
from mindfoundry.optaas.client.sklearn_pipelines.parameter_maker import SklearnParameterMaker
from mindfoundry.optaas.client.sklearn_pipelines.utils import ParametersAndConstraints, SMALLEST_NUMBER_ABOVE_ZERO


class _OptimizableLinearModel(OptimizableBaseEstimator):
    def make_parameters_and_constraints(self, sk: SklearnParameterMaker, **kwargs) -> ParametersAndConstraints:
        """Generates :class:`Parameters <.Parameter>` and :class:`Constraints <.Constraint>` to optimize a linear model."""

        return [sk.FloatParameter('alpha', minimum=SMALLEST_NUMBER_ABOVE_ZERO, maximum=1)], []


class Lasso(BaseLasso, _OptimizableLinearModel):
    """Allows us to optimize a :class:`.Lasso` estimator."""
    pass


class Ridge(BaseRidge, _OptimizableLinearModel):
    """Allows us to optimize a :class:`.Ridge` estimator."""
    pass



from typing import Dict

from mindfoundry.optaas.client.utils import _pprint


class Configuration:
    """A set of values assigned to each :class:`.Parameter`, as generated by OPTaaS or defined by the user.

    Attributes:
        json (Dict): The full Configuration object as stored in OPTaaS in JSON format.
        id (str): Unique id for the Configuration.
        values (Dict): Values assigned to each Parameter.
            Dict keys will be the Parameter names, and values will correspond to the Parameter type.
            :class:`GroupParameters <.GroupParameter>` will appear as nested Dicts.
        type (str): One of

            * `default`: all parameters are set to their default values
            * `exploration`: values are chosen so as to explore a new region of the parameter space
            * `exploitation`: values are chosen so as to improve upon a previously provided result
            * `user-defined`: values are provided by the user
    """

    def __init__(self, json: Dict) -> None:
        self.json = json
        self.id = json['id']  # pylint: disable=invalid-name
        self.values = json['values']
        self.type = json.get('type')
        self.results_url = json['_links']['results']['href']

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.__dict__ == other.__dict__

    def __repr__(self):
        return _pprint(self, 'type', 'values')



import abc
from enum import Enum
from typing import List, Union, Any, Optional, Dict

from mindfoundry.optaas.client.constraint import ConstraintExpression, UnaryPredicate


class Parameter(ConstraintExpression, abc.ABC):
    """Superclass for all parameters.

    Attributes:
        name (str): Human-readable name for the parameter.
        type (str): Parameter type (e.g. "integer" or "choice")
        id (str, optional): Unique id for this Parameter. If not specified, the object id will be used.
        optional (bool, optional, default False): Whether the parameter can be omitted in a :class:`.Configuration`.
        include_in_default (bool, optional, default True): Whether an optional parameter will be included in the :attr:`default Configuration<.Configuration.type>`.
        default (Any, optional): Value to use when generating the :attr:`default Configuration<.Configuration.type>`.
    """

    def __init__(self, name: str, type_: str, id_: Optional[str],
                 optional: Optional[bool] = False, include_in_default: Optional[bool] = True, default=None) -> None:
        self.id = id_ or str(id(self))  # pylint: disable=invalid-name
        self.name = name
        self.type = type_
        if optional:
            self.optional = True
        if default is not None:
            self.default = default
        if include_in_default is not None:
            self.includeInDefault = include_in_default  # pylint: disable=invalid-name

    @abc.abstractmethod
    def is_compatible_value(self, value: Any) -> bool:
        """Returns True iff the given value is compatible with this Parameter."""
        pass

    def is_present(self) -> ConstraintExpression:
        """Returns a :class:`.ConstraintExpression` that evaluates to True if this parameter is present in a :class:`.Configuration`"""
        return UnaryPredicate(self, 'is_present')

    def is_absent(self) -> ConstraintExpression:
        """Returns a :class:`.ConstraintExpression` that evaluates to False if this parameter is present in a :class:`.Configuration`"""
        return UnaryPredicate(self, 'is_absent')

    def to_json(self) -> Dict:
        """JSON representation of this Parameter (used when making a request to OPTaaS to create the :class:`.Task`)"""
        return vars(self)

    def __repr__(self):
        return str(self.to_json())

    def __str__(self):
        return self.name

    def to_optaas_expression(self) -> str:
        return '#' + self.id


class GroupParameter(Parameter):
    """A set of Parameters, all of which will be included in each :class:`.Configuration` unless marked as :attr:`.Parameter.optional`.

    There is no `default` value for groups.

    Attributes:
        items (List[Parameter]): Parameters that comprise this Group. Can be empty.
    """

    def __init__(self, name: str, items: List[Parameter], id: str = None,  # pylint: disable=redefined-builtin
                 optional: bool = None, include_in_default: bool = None) -> None:
        super().__init__(name, 'group', id, optional, include_in_default)
        self.items = items

    def to_json(self) -> dict:
        json = super().to_json().copy()
        json['items'] = [p.to_json() for p in self.items]
        return json

    def is_compatible_value(self, value: Any) -> bool:
        """Always False because GroupParameter never stores a single value."""
        return False


class ChoiceParameter(Parameter):
    """A set of Parameters, only one of which will be chosen to appear in each :class:`.Configuration`.

    Attributes:
        choices (List[Parameter]): Parameters to choose from. Cannot be empty.
        default (Parameter, optional): Default choice, must be present in the `choices` list.
            If not specified, the default will be the first item in `choices`.

    Raises:
        :class:`.ValueError` if `default` is not present in `choices`.
    """

    def __init__(self, name: str, choices: List[Parameter], id: str = None,  # pylint: disable=redefined-builtin
                 optional: bool = None, include_in_default: bool = None, default: Parameter = None) -> None:
        if default is None:
            default_id = None
        else:
            if not any(choice for choice in choices if choice is default):
                raise ValueError(f"Default '{default}' is not one of the choices in ChoiceParameter '{name}'")
            default_id = f'#{default.id}'

        super().__init__(name, 'choice', id, optional, include_in_default, default_id)
        self.choices = choices

    def to_json(self) -> dict:
        json = super().to_json().copy()
        json['choices'] = [c.to_json() for c in self.choices]
        return json

    def is_compatible_value(self, value: Any) -> bool:
        """Always False because ChoiceParameter never stores a single value."""
        return False


class CategoricalParameter(Parameter):
    """A Parameter which can be assigned any value from a specified list of allowed values.

    Args:
        values (List[Union[str, int, float, bool]]): Allowed values (can be string, numeric, boolean or any mixture of those).
        default (Union[str, int, float, bool], optional):
            If defined, must be equal to a value from `values`.
            If not defined, the default will be the first value from `values`.
    """

    def __init__(self, name: str, values: List[Union[str, int, float, bool]],
                 id: str = None,  # pylint: disable=redefined-builtin
                 optional: bool = None, include_in_default: bool = None, default=None) -> None:
        if default is not None and default not in values:
            raise ValueError(f"Default '{default}' is not one of the values in CategoricalParameter '{name}'")
        super().__init__(name, 'categorical', id, optional, include_in_default, default)
        self.enum = values

    def is_compatible_value(self, value: Any) -> bool:
        """Value must be identical to one of the specified values."""
        for allowed_value in self.enum:
            if value == allowed_value and type(value) == type(allowed_value):
                return True
        return False


class SubsetParameter(Parameter):
    """A parameter that may contain any subset of a valid set of values.

    Args:
        values (List[Union[str, int, float, bool]])
            Allowed values (can be string, numeric, boolean or any mixture of those).
        default (List[Union[str, int, float, bool]])
            The default for this kind of parameter must be a subset of allowed values. Optional. Assumed to be the empty
            set if not otherwise specified.

    """

    def __init__(self, name: str, values: List[Union[str, int, float, bool]],
                 id: str = None,  # pylint: disable=redefined-builtin
                 optional: bool = None, include_in_default: bool = None,
                 default: List[Union[str, int, float, bool]] = None) -> None:
        super().__init__(name, 'subset', id, optional, include_in_default, default)
        self.allowedValues = values  # pylint: disable=invalid-name

        if not values:
            raise ValueError(f"The empty list is not a valid set of values for SubsetParameter '{name}'")

        if None in values:
            raise ValueError(f"'None' is not a valid value for SubsetParameter '{name}'")

        if (default is not None) and (not self.is_compatible_value(default)):
            raise ValueError(f"Default '{default}' is not a subset of the values in SubsetParameter '{name}'")

    def is_compatible_value(self, value: List[Any]) -> bool:
        """
        Value must be a subset of allowed values, i.e. all of its element must be included in the set of
        allowed values. An empty subset is also a valid value by extension. No duplicate items are allowed.
        """
        if value is None:
            return False

        value_set = set(value)

        # Fails if either any item is not a value of enum or if item was seen more than once
        return (value_set <= set(self.allowedValues)) and (len(value) == len(value_set))


class ConstantParameter(Parameter):
    """A Parameter which will always be assigned a specified value.

    Args:
        value (Union[str, int, float, bool]): value (can be string, numeric or boolean).
    """

    def __init__(self, name: str, value: Union[str, int, float, bool],
                 id: str = None,  # pylint: disable=redefined-builtin
                 optional: bool = None, include_in_default: bool = None) -> None:
        super().__init__(name, 'constant', id, optional, include_in_default)
        self.value = value

    def is_compatible_value(self, value: Any) -> bool:
        """Value must be identical to the specified value."""
        return value == self.value and type(value) == type(self.value)


class BoolParameter(Parameter):
    """A Parameter which will be assigned either True or False.

    Args:
        default (bool, optional): False if not specified.
    """

    def __init__(self, name: str, id: str = None,  # pylint: disable=redefined-builtin
                 optional: bool = None, include_in_default: bool = None, default: bool = None) -> None:
        if default is not None and not isinstance(default, bool):
            raise ValueError(f"Default '{default}' is not a boolean in BoolParameter '{name}'")
        super().__init__(name, 'boolean', id, optional, include_in_default, default)

    def is_compatible_value(self, value: Any) -> bool:
        """Value must be a bool."""
        return isinstance(value, bool)


class Distribution(Enum):
    """Specifies the distribution of values for a NumericParameter.

    This will influence the values assigned by OPTaaS in each :class:`.Configuration`.
    """
    UNIFORM = "Uniform"
    LOGUNIFORM = "LogUniform"


class NumericParameter(Parameter, abc.ABC):
    """Superclass for FloatParameter and IntParameter"""

    def __init__(self, name: str, type_: str, minimum: Union[int, float], maximum: Union[int, float],
                 id: str = None,  # pylint: disable=redefined-builtin
                 optional: bool = None, include_in_default: bool = None,
                 default: Union[int, float] = None, distribution: Distribution = None) -> None:
        super().__init__(name, type_, id, optional, include_in_default, default)
        self.minimum = minimum
        self.maximum = maximum
        if distribution is not None:
            self.distribution = distribution.value

    @abc.abstractmethod
    def is_compatible_value(self, value: Any) -> bool:
        pass


class IntParameter(NumericParameter):
    """A Parameter which can be assigned any integer value between the `minimum` and `maximum` (inclusive).

    Args:
        minimum (int): Smallest allowed value.
        maximum (int): Largest allowed value.
        default (int, optional):
            If defined, must be a value between `minimum` and `maximum`.
            If not defined, it will be set to the midpoint between `minimum` and `maximum`.
        distribution (Distribution, optional, default Distribution.UNIFORM): See :class:`.Distribution`
    """

    def __init__(self, name: str, minimum: int, maximum: int, id: str = None,  # pylint: disable=redefined-builtin
                 optional: bool = None, include_in_default: bool = None,
                 default: int = None, distribution: Distribution = None) -> None:
        if default is not None and not isinstance(default, int):
            raise ValueError(f"Default '{default}' is not an integer in IntParameter '{name}'")

        super().__init__(name, 'integer', minimum, maximum, id, optional, include_in_default, default, distribution)

    def is_compatible_value(self, value: Any) -> bool:
        """Value must be an int (not bool) and within the range [minimum - maximum] (inclusive)."""
        return isinstance(value, int) and not isinstance(value, bool) and self.minimum <= value <= self.maximum


class FloatParameter(NumericParameter):
    """A Parameter which can be assigned any float value between the `minimum` and `maximum` (inclusive).

    Args:
        minimum (float): Smallest allowed value.
        maximum (float): Largest allowed value.
        default (float, optional):
            If defined, must be a value between `minimum` and `maximum`.
            If not defined, it will be set to the midpoint between `minimum` and `maximum`.
        distribution (Distribution, optional, default Distribution.UNIFORM): See :class:`.Distribution`
        cyclical (bool, optional, default False):
            If True, OPTaaS will select values from a period starting at the `minimum` (inclusive) and ending at the `maximum`
            (exclusive). Values near the minimum and maximum will be considered to be close, as if they were on a circle.
            **Note:** if `cyclical` is true, `distribution` will be ignored. Also, if any of your parameters are cyclical, all
            your parameters must be Floats, Constants or Groups (other types are not currently supported), and none of them can
            be `optional`.
    """

    def __init__(self, name: str, minimum: float, maximum: float, id: str = None,  # pylint: disable=redefined-builtin
                 optional: bool = None, include_in_default: bool = None, default: float = None,
                 distribution: Distribution = None, cyclical: bool = None) -> None:
        if default is not None and not isinstance(default, float) and not isinstance(default, int):
            raise ValueError(f"Default '{default}' is not a float in FloatParameter '{name}'")

        super().__init__(name, 'number', minimum, maximum, id, optional, include_in_default, default, distribution)
        if cyclical is not None:
            self.cyclical = cyclical

    def is_compatible_value(self, value: Any) -> bool:
        """Value must be a float or int (not bool) and within the range [minimum - maximum] (inclusive)."""
        return isinstance(value, (float, int)) and not isinstance(value, bool) and self.minimum <= value <= self.maximum
